/*
 *   ----------------------------------------------
 *   r e D A T E   -   the file and archive redater
 *   ----------------------------------------------
 *
 *   reDATE  -  File and Archive Redater  -  V1.0
 *   Copyright (c) 1988  -  Dean Tutterow  -  All Rights Reserved
 *   
 *   What does it do?   Why it allows you to redate files of course.
 *   How does it do it?  Why easily, quickly, and just like you tell
 *   it.   This program was written simply because I couldn't find a
 *   single  program  out there in BBS-land that WORKed properly all 
 *   of the time including one popular one that has made the rounds.
 *   There are no errors included as features as far as I know.
 *   
 *   USAGE:
 *   
 *   reDATE  tgtfile [/options]
 *
 *	 where [/options] are:
 *		/ARC		  reDATE archive file to latest contained file
 *		/DATE[=mm-dd-yy]  specify date for file
 *		/HELP   	  display usage
 *		/LOG		  log reDATE actions
 *		/TEST		  show what would happen, but don't do it
 *		/TIME[=hh:mm:ss]  specify time for file
 *		/TODAY		  set file date/time to now
 *   
 *   The tgtfile should be a valid DOS filename; wildcards are accepted.
 *   You need only supply enough  to the  option  name to distinguish it
 *   from the other options, for example /A and /TE.
 *   
 *   
 *   Typical command lines are:
 *   
 *   reDATE C:\LIBRARY\*.ARC /ARC --> redates all '.ARC' files in C:\LIBRARY
 *   				      to the latest file in that archive
 *   reDATE E:\BOOKS.WKS/TODAY	  --> redate file to the current date/time
 *   reDATE BOOKS.WKS/TI=22:00	  --> redate file to 10:00 PM
 *   
 *   This program was developed with MICROSOFT C v4.0 (antique but it is
 *   all I have).
 *      MSC /Zp REDATE;
 *      LINK REDATE;
 */

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <string.h>
#include <time.h>
#include <types.h>
#include <timeb.h>

#define	DEBUG(x) if (debug) {x}
#define DEBUG_COMMAND_LINE \
    if (debug) {\
	fprintf (stdout, "\nCommand line contains:\t%s", equivalence);\
	for (i = 0; i < value; i++) {\
            fprintf (stdout,"\nParsed value %d contains:\t%s", i, elements[i]);\
            status = parse_option (elements[i], &option);\
            if (status) fprintf (stdout,"\n                options:\t%s", option);\
	}\
    }
#define byte	unsigned char
#define word	unsigned int
#define MAX_NUM_OF_ELEMENTS 6

/*  E R R O R   C O D E S  */
#define UNKNOWN             0x00
#define FILENAME_TROUBLE    0x01
#define INVALID_OPTION      0x02
#define INVALID_ARCHIVE     0x03

/*  S E V E R I T Y   C O D E S  */
#define	FATAL		"F"
#define	SEVERE		"S"
#define	WARNING		"W"
#define INFORMATIONAL	"I"

/*  G L O B A L   D A T A  */ 
byte    ARC = 0;                /* if set, file is archive      */
word    DATE = 0;               /* date to stamp file to        */
word    TIME = 0;               /* time to stamp file to        */
byte    debug = 0;              /* if set, prints DEBUG lines   */
byte    HELP = 0;               /* if set, print usage info     */
byte    LOG = 0;                /* if set, logs reDATE actions  */
byte    TEST = 0;               /* if set, show what would occur*/
int     TODAY = 0;              /* if set, set d/t stamp to now */
char    *INFILE;                /* pointer to input filename    */
char    MONIKER[80];            /* storage for find-file name   */
char    *findfile();            /* findfile() returns char ptr  */
char    hdrver;                 /* header version               */
FILE    *file;                  /* the old file                 */
char    filename[100];          /* storage for filename         */
word    filedate;               /* file date stamp              */
word    filetime;               /* file archive time stamp      */
word    olddate = 0;            /* old file date stamp          */
word    oldtime = 0;            /* old file archive time stamp  */

static  struct {
        char dummy[21];         /* reserved for dos             */
        byte attribute;         /* returned attribute           */
        word time;              /* file time stamp              */
        word date;              /* file date stamp              */
        long size;              /* size of file                 */
        char fn[13];            /* filename string              */
} ff_area;

struct  heads {                 /* archive entry header format  */
        char name[13];          /* file name                    */
        long size;              /* size of file, in bytes       */
        word date;              /* creation date                */
        word time;              /* creation time                */
        int crc;                /* cyclic redundancy check      */
        long length;            /* true file length             */
};



main    (argc, argv)

int     argc;
char    *argv[];
{
    int     i, value;
    char    equivalence[100], parsed[120], *elements[MAX_NUM_OF_ELEMENTS];
    char    *d, *option, *parse_option(), *status, *file_status;
    static  char    *index1, *index2;

    print_copyright();
    memset (equivalence, '\0', sizeof (equivalence));
    memset (parsed, '\0', sizeof (parsed));

    for (i = 0; i < argc; i++) {
        strcat (equivalence, argv[i]);
        strcat (equivalence, " "); }

    value = MAX_NUM_OF_ELEMENTS;
    parse_command_line (equivalence, parsed, elements, &value);
    get_filenames_and_options (elements, &value, &INFILE);
    DEBUG_COMMAND_LINE;
    if (HELP || !INFILE) { print_usage (); return; }
    if (TEST)
        fprintf (stdout,
	    "\nHere is how it would look IF we were not just testing!\n");

    fixpath (&index1, &index2);
    for ( d=findfile(INFILE,0); *d; d=findfile(NULL,0) ) {
        strcpy (index1, d);
        if( !(file = fopen(filename,"rb")) )
	    err_exit (FILENAME_TROUBLE, FATAL);
        getstamp (file,&olddate,&oldtime);  /* get a file's date/time stamp */
        redate (file);                      /* redate file                  */
        fclose (file);                      /* close the file               */
    }
    exit (0);
}


int	print_copyright ()
{
    fprintf (stderr, "reDATE  -  File and Archive Redater  -  V1.0\n");
    fprintf (stderr, "Copyright (c) 1988  -  Dean Tutterow  -  All Rights Reserved\n");
}

int 	_nullcheck () {};
int 	_setenvp () {};


int     redate (file)                       /* redate file                  */

FILE    *file;                              /* the file                     */
{
    struct  heads arch;                     /* target archive header        */
    int     status;                         /* has an error been detected   */

    if (ARC) {
        filedate = filetime = 0;            /* set to really low values     */
        while (status = readhdr(&arch,file)) {
            if (status == -1) break;
            if(arch.date>filedate ||
              (arch.date==filedate && arch.time>filetime)) {
                filedate = arch.date;
                filetime = arch.time;
            }                               /* else report on all files     */
            fseek(file,arch.size,1);        /* skip to next header          */
        }
        if (status != -1) setstamp (file, filedate, filetime);
    }
    else
        setstamp (file, DATE, TIME);
}


int     getstamp (f, date, time)            /* get a file's date/time stamp */

FILE    *f;                                 /* file to get stamp from       */
word    *date, *time;                       /* storage for the stamp        */
{
    union REGS reg;

    reg.x.ax = 0x5700;                      /* get date/time                */
    reg.x.bx = fileno(f);                   /* file handle                  */
    intdos (&reg,&reg);                     /* DOS call                     */
    if(reg.x.cflag)                         /* test DOS call                */
        err_exit (UNKNOWN, WARNING);
    *date = reg.x.dx;                       /* save date/time               */
    *time = reg.x.cx;
    date_time_to_string (reg.x.dx, reg.x.cx);
}

int     setstamp (f, date, time)            /* set a file's date/time stamp */

FILE    *f;                                 /* file to set stamp on         */
word    date, time;                         /* desired date, time           */
{
    union REGS reg;

    if (olddate == date && oldtime == time) {
        if (LOG)  fprintf (stdout, "  no change"); return; }
    if (LOG) fprintf (stdout, "  changed to  ");
    date_time_to_string (date, time);
    if (TEST) return;                       /* just go through the motions  */
    reg.x.ax = 0x5701;                      /* set date/time                */
    reg.x.bx = fileno(f);                   /* file handle                  */
    reg.x.cx = time;                        /* desired time                 */
    reg.x.dx = date;                        /* desired date                 */
    intdos (&reg,&reg);                     /* DOS call                     */
    if(reg.x.cflag)                         /* test DOS call                */
        err_exit (UNKNOWN, WARNING);
}



int     readhdr (hdr,arc)                   /* read a header from an archive */

struct  heads *hdr;                         /* storage for header            */
FILE    *arc;                               /* archive to read header from   */
{
    char    name[13];                       /* filename buffer               */
    int     try = 0;                        /* retry counter                 */
    static  int first = 1;                  /* true only on first read       */

    if (!arc || feof(arc))                  /* if archive didn't open or no  */
        return 0;                           /*   more data then pretend it's */
                                            /*   the end                     */
    if (fgetc(arc)!=26) {                   /* check archive validity        */
        err_exit (INVALID_ARCHIVE, WARNING);
        return -1;
    }
    hdrver = fgetc(arc);                    /* get header version            */
    if (hdrver<0) {
        err_exit (INVALID_ARCHIVE, WARNING);
        return -1;
    }
    if (hdrver==0)
        return 0;                           /* note our end of archive marker*/
    fread(hdr,sizeof(struct heads),1,arc);
    first = 0; return 1;                    /* we read something             */
}



char    *findfile (filename, mode)          /* get files, one by one         */

char    *filename;                          /* template, or NULL             */
int     mode;                               /* search mode bits              */
{
    union   REGS reg;
    static  int  first = 1;                 /* true only on first call       */
    static  char *result = MONIKER;

    segread(&reg.x.si);                     /* get ds value                  */
    if(filename) {                          /* if filename is given          */
        reg.x.dx = (int) &filename[0];      /* then use it                   */
        reg.x.ax = 0x4e00;                  /* and search for first          */
    }
    else if(first)                          /* if no name and first call     */
        return NULL;                        /* then not much we can do       */
    else reg.x.ax = 0x4f00;                 /* else search for next          */
    first = 0;                              /* no longer first time          */

    reg.x.cx = mode;                        /* set search modes              */
    bdos(0x1a,&ff_area);                    /* set the transfer address      */
    intdos (&reg,&reg);                     /* DOS call                      */
    if(reg.x.cflag)                         /* test DOS call                 */
        return NULL;                        /* no more files                 */
    strcpy(result,ff_area.fn);              /* save name of file             */
    if (LOG) fprintf (stdout, "\n%-14s", ff_area.fn);
    return result;
}



int	convert_date_time (string, date, time)

char	*string;
word	*date;
word 	*time;
{
    int     year, month, day, hour, min, sec;
    char    *ytok, *mtok, *dtok, *htok, *mintok, *stok, empty_string[2];

    strcpy (empty_string, "\0");
    if (date) {
        mtok = strtok (string, "-:");         /* Month token */
        if (mtok == NULL) mtok = empty_string;
        dtok = strtok (NULL, "-:");           /* Date token  */
        if (dtok == NULL) dtok = empty_string;
        ytok = strtok (NULL, "-:");           /* Year token  */
        if (ytok == NULL) ytok = empty_string;
        year  = atoi (ytok);
        if (year < 80) year = 0; else year = year-80; year = year<<9;
        month = atoi (mtok); month = month<<5;
        day   = atoi (dtok);
        *date = (year + month + day);
    }
    if (time) {
        htok = strtok (string, "-:");         /* Hours token   */
        if (htok == NULL) htok = empty_string;
        mintok = strtok (NULL, "-:");         /* Minutes token */
        if (mintok == NULL) mintok = empty_string;
        stok = strtok (NULL, "-:\0");         /* Seconds token */
        if (stok == NULL) stok = empty_string;
        hour  = atoi (htok);   hour = hour<<11;
        min   = atoi (mintok); min = min<<5;
        sec   = atoi (stok);
        *time = (hour + min + sec/2);
    }
}


int 	date_time_to_string (date, time)

word	date;
word 	time;
{
    int     year, month, day, hour, min, sec;

    if (LOG) {
	year  = (date >> 9) & 0x7f;      /* dissect the date */
	month = (date >> 5) & 0x0f;
	day   = date & 0x1f;

	hour  = (time >> 11) & 0x1f;     /* dissect the time */
	min   = (time >> 5) & 0x3f;
	sec   = (time & 0x1f) * 2;

	fprintf (stdout, "%02d-%02d-%02d %02d:%02d:%02d",\
			 month, day, (year+80), hour, min, sec);
    }
    return;
}	



int	parse_command_line (input, output, arrayed, number)

char	*input, *output, *arrayed[];
int	*number;
{
    int     i, j, k, paren;

    /* Convert input line to upper case */
    for (i=0; i < strlen(input); i++)  (input)[i] = toupper ( (input)[i] );
    (output)[0] = ' ';

    for (i=k=paren=0, j=1; (i < strlen (input) && k <= *number); i++) {
        switch ( (input)[i] ) {
            case ','  : if (paren) { (output)[j] = (input)[i]; break; }
            case ' '  :
            case '\t' : (output)[j] = ' '; break;
            case '/'  : (output)[j] = ' '; j++; paren--;
            case '('  : paren += 2;
            case ')'  : paren--;
            default   : (output)[j] = (input)[i];
        }
        j++;
        /*
         * If this is the start of an argument, then place its address
         * in the argument list & null-terminate the previous argument.
         */
        if ( (output)[j-1] != ' '  &&  (output)[j-2] == ' ' ) {
            if (k != *number) {
                (arrayed)[k] = (&(output)[j-1]);
                k++;
            }
            (output)[j-2] = '\0';
        }
    }
    (output)[j] = '\0';
    *number = k;
};



char	*parse_option (option, output)

char	*option, **output;
{
    int length_to_option;

    if (*option != '/') return (0);
    if (!(length_to_option = strcspn (option, "="))) return (0);
    if (length_to_option == strlen (option)) return (0);
    *output = option + length_to_option + 1;
    return ((char *)output);
}



int	get_filenames_and_options (elements, value, INFILE)

char	**INFILE, *elements[];
int	*value;
{
    int     first, i, j;
    long    timestore;
    char    *status, *option, *parse_option(), datetime[15];
    char    monthstring[40], *timeptr, char_1, char_2;
    struct  timeb   timebuffer;

    *INFILE = 0;
    strcpy (monthstring, "JANFEBMARAPRMAYJUNJULAUGSEPOCTNOVDEC");
    for (i = first = 1; i < *value; i++) {
        if (first && *elements[i] != '/') {
	    *INFILE = elements[i]; first = 0; } 
        else {
	    char_1 = *(elements[i] + 1); char_2 = *(elements[i] + 2);
	    switch (char_1) {
		case 'A' : ARC = 1; break;
		case 'D' :
                    switch (char_2) {
                        case 'A' :
                            status = parse_option (elements[i], &option);
                            if (status) {
                                sscanf (option, "%14s", datetime);
                                convert_date_time (datetime, &DATE, 0); }
                            break;
                        case 'E' : debug = LOG = 1; break;
                    }
                    break;
		case 'H' : case '?' : HELP = 1; break;
		case 'L' : LOG  = 1; break;
		case 'T' :
                    switch (char_2) {
                        case 'E' : TEST = LOG = 1; break;
                        case 'I' :
                            status = parse_option (elements[i], &option);
                            if (status) {
                                sscanf (option, "%s", datetime);
                                convert_date_time (datetime, 0, &TIME); }
                            break;
                        case 'O' :
                            ftime (&timebuffer);
                            timeptr = ctime (&timebuffer.time);
                            strupr (timeptr);
                            j = index (monthstring, timeptr+4, 3);
                            j = (j + 3) / 3;
                            memset (datetime, '\0', sizeof (datetime));
                            itoa (j, datetime, 10);
                            strcat (datetime, "-");
                            strncat (datetime, timeptr+8, 2);
                            strcat (datetime, "-");
                            strncat (datetime, timeptr+22, 2);
                            convert_date_time (datetime, &DATE, 0);
                            memset (datetime, '\0', sizeof (datetime));
                            strncpy (datetime, timeptr+11, 8);
                            convert_date_time (datetime, 0, &TIME);
                            TODAY = 1;
                            break;
                    }
                    break;
		default : 
                    err_exit (INVALID_OPTION, WARNING);
	    }
        }
    }
}


int 	index (s, t, c)

char	*s, *t;
int	c;
{
    int      i;
    register int  j, k;

    for (i = 0; s[i] != '\0'; i++) {
        for (j = i, k = 0; t[k] != '\0' && s[j] == t[k] && k < c; j++, k++)
	    ;
        if (t[k] == '\0' || k == c)
	    return (i);
    }
    return (-1);
}



int	fixpath (index)

char	**index;
{
    static  char    *indexa, *indexb;

    strcpy (filename,INFILE);       /* get ready to fix path */
    indexa = strrchr (filename, ':');
    if (!indexa) indexb = strrchr (filename, '\\');
    else indexb = strrchr (indexa, '\\');
    if (!indexa && !indexb) indexa = filename;
    else {
        if (!indexa) indexa = indexb;
        indexa++; }
    *index = indexa;
}

  	

int	print_usage() 

{
    fprintf (stderr, "\nusage: reDATE tgtfile [/options]\n\n");
    fprintf (stderr, " where [/options] are:\n");
    fprintf (stderr, "\t/ARC");
    fprintf (stderr, "\t\t  reDATE archive file to latest contained file\n");
    fprintf (stderr, "\t/DATE[=mm-dd-yy]  specify date for file\n");
    fprintf (stderr, "\t/DEBUG\n");
    fprintf (stderr, "\t/HELP   \t  display this!\n");
    fprintf (stderr, "\t/LOG\t\t  log reDATE actions\n");
    fprintf (stderr, "\t/TEST");
    fprintf (stderr, "\t\t  show what would happen, but don't do it\n");
    fprintf (stderr, "\t/TIME[=hh:mm:ss]  specify time for file\n");
    fprintf (stderr, "\t/TODAY\t\t  set file date/time to now\n");
    return;
}



int	err_exit (code, severity)

int	code;
char	*severity;
{
    char    err_code[20], msg[80];

    /* Display an error message and exit if appropriate. */
    switch (code) {

        case FILENAME_TROUBLE :
            strcpy (msg, "Trouble opening file, please check filename.");
            strcpy (err_code, "FILEOPEN");
            break;

        case INVALID_OPTION :
            strcpy (msg, "You have specified an invalid option.");
            strcpy (err_code, "INVLDOPT");
            break;

        case INVALID_ARCHIVE :
            strcpy (msg, "An invalid or unknown '.arc' format has been detected.");
            strcpy (err_code, "INVLDARC");
            break;

        default :
            strcpy (msg, "Unknown error!");
            strcpy (err_code, "UNKNOWN");
	}
    fprintf (stderr, "\nreDATE-%c-%s, %s", *severity, err_code, msg);
    if ( (*severity == 'F')  ||  (*severity == 'S') ) {
        print_usage();
        exit (1);             /* Exit with errorlevel 1 */
    }
    return;
}
